from typing import Union

from treepath.path.exceptions.path_syntax_error import PathSyntaxError
from treepath.path.traverser.key_match import KeyMatch
from treepath.path.traverser.list_match import ListMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class CatchState:
    __slots__ = 'iterable', \
                'match_constructor'

    def __init__(self, iterable, match_constructor):
        self.iterable = iterable
        self.match_constructor = match_constructor


class TupleVertex(Vertex):
    __slots__ = '_tuple'

    def __init__(self, parent, tuple_: tuple):
        self._tuple = tuple_
        super().__init__(parent, tuple_)
        self.is_catch_vertex = True
        self.validate_tuple()

    @property
    def path_segment(self):
        tuple_ = self._tuple
        return f"[{', '.join(map(repr, tuple_))}]"

    def validate_tuple(self):
        for value in self._tuple:
            if not (isinstance(value, int) or isinstance(value, str)):
                raise PathSyntaxError(self.parent,
                                      f"Unsupported indices: {self.path_segment}.  Comma serrated indices can be "
                                      f"a combination of int and str values.",
                                      self.path_segment)

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:

        remembered_catch_state = parent_match.remembered_catch_state
        if not remembered_catch_state:
            remembered_catch_state = self._create_on_catch_state(parent_match)
            if not remembered_catch_state:
                return None
            traverser.remember_on_catch(parent_match, remembered_catch_state)

        try:
            item = next(remembered_catch_state.iterable)
            return remembered_catch_state.match_constructor(
                parent_match,
                item[0],
                item[1],
                self,
                vertex_index,
                parent_match.remembered_on_catch_match,
                parent_match.remembered_on_catch_action
            )
        except StopIteration:
            traverser.restore_on_catch(parent_match)
            return None

    def _create_on_catch_state(self, parent_match: TraverserMatch):
        data = parent_match.data
        if isinstance(data, dict):
            return self._key_match(parent_match)
        elif isinstance(data, list):
            return self._list_match(parent_match)
        else:
            return None

    def _key_match(self, parent_match: TraverserMatch) -> CatchState:
        tuple__ = self._tuple

        def my_iter():
            tuple_ = tuple__
            data = parent_match.data
            for key in tuple_:
                try:
                    value = data[key]
                    yield key, value
                except KeyError:
                    continue

        iterable = my_iter()
        return CatchState(iterable, KeyMatch)

    def _list_match(self, parent_match: TraverserMatch) -> CatchState:
        tuple__ = self._tuple

        def my_iter():
            tuple_ = tuple__
            data = parent_match.data
            for index in tuple_:
                try:
                    value = data[index]
                    yield index, value
                except (IndexError, TypeError):
                    continue

        iterable = my_iter()
        return CatchState(iterable, ListMatch)



"""
See README.md  for details
"""
from treepath.path.builder.dash_path_builder import DashPathRoot
from treepath.path.builder.patch_constants import rec
from treepath.path.builder.patch_constants import recursive
from treepath.path.builder.patch_constants import wc
from treepath.path.builder.patch_constants import wildcard
from treepath.path.builder.root_path_builder import RootPathBuilder
from treepath.path.exceptions.infinite_loop_detected import InfiniteLoopDetected
from treepath.path.exceptions.match_not_found_error import MatchNotFoundError
from treepath.path.exceptions.nested_match_not_found_error import NestedMatchNotFoundError
from treepath.path.exceptions.path_syntax_error import PathSyntaxError
from treepath.path.exceptions.stop_traversing import StopTraversing
from treepath.path.exceptions.traversing_error import TraversingError
from treepath.path.exceptions.treepath_exception import TreepathException
from treepath.path.traverser.match import Match
from treepath.path.traverser.trace import Trace
from treepath.path.traverser.trace import log_to
from treepath.path.traverser.traverser_functions import find
from treepath.path.traverser.traverser_functions import find_matches
from treepath.path.traverser.traverser_functions import get
from treepath.path.traverser.traverser_functions import get_match
from treepath.path.traverser.traverser_functions import has
from treepath.path.traverser.traverser_functions import has_all
from treepath.path.traverser.traverser_functions import has_any
from treepath.path.traverser.traverser_functions import has_not
from treepath.path.traverser.traverser_functions import nested_find_matches
from treepath.path.traverser.traverser_functions import nested_get_match

# path is a global object for dynamically declaring a query to extract data from a json data structure.
#     # A treepath example that fetches the value 1 from data.
#     data = {
#         "a": {
#             "b": [
#                 {
#                     "c": 1
#                 },
#                 {
#                     "c": 2
#                 }]
#         }
#     }
#     value = get(path.a.b[0].c, data)
#     assert value == 1
#
#     # A treepath example that fetches the values 1 and 2 from data.
#     value = [value for value in find(path.a.b[wc].c, data)]
#     assert value == [1,2]
# See the README.md for example on how to use the path object.
path = RootPathBuilder()

# pathd is similar to path, but with the augmentation to translate underscores into dashes.  It provides a convenient
# way of working with json documents that has a lot of keys with dashes in the names as dashes are not valid python
# attribute names.  For example pathd.a_a  is equivalent to data["a-a"]
pathd = DashPathRoot()



import os
from typing import Union

from treepath.path.exceptions.treepath_exception import TreepathException
from treepath.path.vertex.vertex import Vertex


class PathSyntaxError(TreepathException, SyntaxError):
    """PathSyntaxError is raised when there is a syntax issue with the path"""

    def __init__(self, parent_vertex: Union[Vertex, None], error_msg, invalid_path_segment):
        self.error_msg = error_msg
        self.invalid_path_segment = invalid_path_segment
        super().__init__(parent_vertex)

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"PathSyntaxError({self.error_msg}{os.linesep}  path: {path}{self.invalid_path_segment})"



from typing import Callable


def pretty_repr(pretty_rep_: Callable[[], str]):
    class PrettyRep:
        __slots__ = '_operator', '_pretty_rep'

        def __init__(self, operator_):
            self._operator = operator_
            self._pretty_rep = pretty_rep_

        def __call__(self, *args):
            return self._operator(*args)

        def __repr__(self):
            return self._pretty_rep()

        def __str__(self):
            return self._pretty_rep()

    return PrettyRep


def add_attr(attr_name, attr_value):
    def _add_attr(func):
        setattr(func, attr_name, attr_value)
        return func

    return _add_attr



import typing

from treepath.path.exceptions.traversing_error import TraversingError
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class RootVertex(Vertex):
    __slots__ = ()

    def __init__(self, name: str):
        super().__init__(None, name)
        self._path_as_list = [self]

    @property
    def path_as_list(self) -> list:
        return self._path_as_list

    @property
    def path(self) -> str:
        return self.name

    def traverse(self, visit: typing.Callable[[Vertex], None]):
        visit(self)

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> typing.Union[TraverserMatch, None]:
        raise TraversingError(parent_match, f"Possible Bug.  The root match method should never be called.")



from treepath.path.exceptions.treepath_exception import TreepathException


class MatchNotFoundError(TreepathException, LookupError):
    """
    MatchNotFoundError is raised when get_match query fails to generate a result for the specified path.
    """

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"MatchNotFoundError(No get_match occurred on path: {path})"



from treepath.path.builder.path_builder import PathBuilder
from treepath.path.vertex.root_vertex import RootVertex


class RootPathBuilder(PathBuilder):
    __slots__ = ()

    def __init__(self):
        vertex = RootVertex("$")
        super().__init__(vertex)



import typing

from treepath.path.traverser.traverser_match import TraverserMatch


class ImaginaryMatch(TraverserMatch):
    """
    An ImaginaryMatch purpose defines additional traversing state without moving the pointer to the vertex in the data tree
    structure.  The traversing algorithms use this object to assist in traversing the data tree structure without
    mangling the real_parent traversing state.
    """
    __slots__ = ()

    @property
    def path_as_list(self) -> list:
        return self.real_parent.path_as_list

    @property
    def data_name(self) -> str:
        return self.real_parent.data_name

    @property
    def path(self) -> str:
        return self.real_parent.path

    @property
    def path_segment(self) -> str:
        return self.real_parent.path_segment

    @property
    def parent(self) -> TraverserMatch:
        return self.real_parent.parent

    def traverse(self, visit: typing.Callable):
        """
        Skip to the real_parent.  This get_match has
        """
        self.real_parent.traverse(visit)



import os

from treepath.path.exceptions.treepath_exception import TreepathException
from treepath.path.traverser.match import Match
from treepath.path.traverser.traverser_match import TraverserMatch


class TraversingError(TreepathException, RuntimeError):
    """TraversingError is raised when path expression cannot traverse the data structure"""

    def __init__(self, match: TraverserMatch, error_msg):
        self.match = Match(match)
        self.error_msg = error_msg
        super().__init__(match.vertex)

    def _resolve_msg(self):
        path = repr(self.vertex)
        match = repr(self.match)
        return f"""TraversingError({self.error_msg}{os.linesep}  path: {path}{os.linesep}  last_match: {match})"""



from typing import Union

from treepath.path.exceptions.traversing_error import TraversingError
from treepath.path.exceptions.treepath_exception import TreepathException
from treepath.path.traverser.imaginary_match import ImaginaryMatch
from treepath.path.traverser.predicate_match import PredicateMatch
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class PredicateVertex(Vertex):
    __slots__ = '_predicate'

    def __init__(self, parent, predicate):
        self._predicate = predicate
        super().__init__(parent, predicate)

    @property
    def path_segment(self):
        return f"[{self._predicate}]"

    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        try:
            if self._predicate(PredicateMatch(traverser.trace, parent_match)):
                match = ImaginaryMatch(
                    parent_match,
                    "PredicateMatch",
                    parent_match.data,
                    self,
                    vertex_index,
                    parent_match.remembered_on_catch_match,
                    parent_match.remembered_on_catch_action
                )
                return match
        except Exception as e:
            error_message = f"Evaluation of predicate failed because of error: {repr(e)}"
            raise TraversingError(parent_match, error_message) from e



from typing import Callable

from treepath.path.traverser.imaginary_match import ImaginaryMatch
from treepath.path.traverser.match_traverser import MatchTraverser
from treepath.path.traverser.trace import Trace
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class NestedMatchTraverser(MatchTraverser):
    __slots__ = 'nested_match'

    def __init__(self,
                 nested_match: TraverserMatch,
                 leaf_vertex: Vertex,
                 trace: Callable[[Trace], None] = None
                 ):
        super().__init__(nested_match.data, leaf_vertex, trace=trace)
        self.nested_match = nested_match

    def init_action(self):
        vertex = self.vertex_path[0]
        root_match = ImaginaryMatch(
            self.nested_match,  # real_parent
            vertex.name,
            self.root_data,  # data
            vertex,
            0,  # vertex_index,
            None,  # remembered_on_catch_match
            self.done_action
        )
        root_match.remembered_on_catch_match = root_match
        self.root_match = root_match
        self.current_match = root_match
        self._invoke_next_action = self.report_action



from treepath.path.traverser.traverser_match import TraverserMatch


class ParentMatch(TraverserMatch):
    __slots__ = '_remembered_parent'

    def __init__(self, remembered_parent, *args):
        super().__init__(*args)
        self._remembered_parent = remembered_parent

    @property
    def path_segment(self) -> str:
        return f"<-{self.real_data_name}"

    @property
    def remembered_parent(self):
        return self._remembered_parent.parent



from abc import ABC, abstractmethod
from typing import Union, Callable

from treepath.path.traverser.traverser_match import TraverserMatch


class Vertex(ABC):
    __slots__ = 'parent', \
                'name', \
                'is_catch_vertex', \
                '_path_as_list', \
                '_path'

    def __init__(self, parent, name):
        self.parent = parent
        self.name = name
        self.is_catch_vertex = False
        self._path_as_list = self
        self._path = self

    @property
    def path_as_list(self) -> list:
        path_as_list = self._path_as_list
        if path_as_list != self:
            return path_as_list

        path_as_list = []

        def collect(vertex):
            path_as_list.append(vertex)

        self.traverse(collect)

        self._path_as_list = path_as_list

        return path_as_list

    @property
    def path(self):
        path = self._path
        if path != self:
            return path

        path_as_list = self.path_as_list
        path = ''.join(vertex.path_segment for vertex in path_as_list)
        self._path = path
        return path

    @property
    def path_segment(self):
        return self.name

    def traverse(self, visit: Callable):
        self.parent.traverse(visit)
        visit(self)

    @abstractmethod
    def match(self, parent_match: TraverserMatch, traverser, vertex_index: int) -> Union[TraverserMatch, None]:
        raise NotImplementedError

    def __repr__(self):
        return self.path

    def __str__(self):
        return self.__repr__()



from treepath.path.exceptions.traversing_error import TraversingError


class InfiniteLoopDetected(TraversingError):
    """
    InfiniteLoopDetected is raised if a infinite loop is detected while traversing the tree searching for the path.
    This uselessly occurs if the next(...) takes excessive amount of time to return.  Most likely the cause is the
    data-structure is not a tree but is actually a graph with a cyclical dependency.
    """

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"InfiniteLoopDetected(Traversing seems to go on for ever on path: {path})"
